/*                                                                  |- CFSFC -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is Copyright (C) 1998 Daniel Berrange.
 * All Rights Reserved. 
 *
 * The Initial Developer of the Original Code is
 * Daniel Berrange.
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * peer.c
 * -------
 *
 * Demo that comms system
 */

#include <stdio.h>
#include <unistd.h>

#include "impulse/util/main.h"
#include "impulse/util/error.h"
#include "impulse/util/memory.h"
#include "impulse/comms/main.h"
#include "impulse/comms/channel.h"

#include <errno.h>

impulse_bool channel_event_handler(void *data)
{
  impulse_channel channel = (impulse_channel)data;
  impulse_unistring sender = NULL;
  impulse_char buffer[1024];
  impulse_char *send = NULL;
  impulse_int got = 0;
  impulse_int pending = 0;
  
  pending = impulse_channel_pending(channel);
  if (pending > 1024)
    pending = 1024;
  got = impulse_channel_receive(channel, (impulse_ubyte *)buffer, pending, &sender);
  
  if (got < 0)
  {
    printf("Error when receiving data\n");
    impulse_quit();
    return 0;
  }
      
  send = impulse_unistring_to_utf(sender, 8, 0);
  printf("Received message from %s (%d bytes)\n", send, got);
  printf("{ %s }\n", buffer);
  impulse_free(send);
  impulse_unistring_destroy(sender);

  return 1;
}


impulse_bool console_event_handler(impulse_event *event,
				   void *data)
{  
  impulse_channel channel = (impulse_channel)data;
  impulse_char mesg[1024];
  impulse_char dest[1024];
  impulse_bool addr = 1;
  impulse_int offset = 0;

  for(;;)
  {
    char c;
    
    if (read(event->data.fd.fd, &c, 1) <= 0)
    {
      printf("Error while reading from console\n");
      impulse_quit();
      return 0;
    }
    
    if (c == ' ' && addr)
    {
      dest[offset++] = '\0';
      offset = 0;
      addr = 0;
      continue;
    }
    
    if (addr)
      dest[offset++] = c;
    else
      mesg[offset++] = c;
    
    if (offset == 1022)
    {
      c = '\n';
      if (addr)
	dest[offset++] = c;
      else
	mesg[offset++] = c;
    }
    
    if (c == '\n')
    {
      impulse_unistring to = NULL;
      if (addr)
      {
	printf("syntax: address message\n");
	return 1;
      }
      
      mesg[offset-1] = '\0';
      to = impulse_unistring_create_utf(dest, 8);
      if (impulse_channel_send(channel, (impulse_ubyte*)mesg, strlen(mesg)+1, to) < 0)
      {
	printf("Could not send message to host %d\n", errno);
	impulse_quit();
	impulse_unistring_destroy(to);
	return 0;
      }
      impulse_unistring_destroy(to);
      break;
    }
  }
  return 1;
}


impulse_int main (impulse_int argc,
		  impulse_char *argv[])
{
  impulse_channel channel = NULL;
  impulse_transport transport = NULL;
  impulse_unistring local;
  impulse_unistring remote;
  impulse_unistring help;
  impulse_unistring arg;
  impulse_command_value value_local = NULL;
  impulse_command_value value_remote = NULL;
  impulse_bool connect = 0;
  impulse_int caps;
  
  impulse_init(argc, argv);
  
  local = impulse_unistring_create_utf("local", 8);
  help = impulse_unistring_create_utf("bind to local address [name]", 8);
  arg = impulse_unistring_create_utf("name", 8);
  impulse_command_option_arg(local, 0, arg, IMPULSE_COMMAND_TYPE_STRING, help);
  impulse_unistring_destroy(help);
  
  remote = impulse_unistring_create_utf("remote", 8);
  help = impulse_unistring_create_utf("connect to remote address [name]", 8);
  impulse_command_option_arg(remote, 0, arg, IMPULSE_COMMAND_TYPE_STRING, help);
  impulse_unistring_destroy(help);
  impulse_unistring_destroy(arg);

  impulse_command_parse();
  impulse_comms_init();
  
  if (impulse_command_set(remote, &value_remote))
    connect = 1;
  
  if (!impulse_command_set(local, &value_local))
    impulse_command_help();
  
  impulse_unistring_destroy(local);
  impulse_unistring_destroy(remote);
  
  if (!(transport = impulse_comms_lookup(value_local->s)))
  {
    impulse_unistring desc = impulse_unistring_create_utf("Unknown transport protocol requested", 8);
    impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, value_local->s);
  }

  caps = IMPULSE_TRANSPORT_PEER | IMPULSE_TRANSPORT_DATAGRAM | IMPULSE_TRANSPORT_CONNECTIONLESS;
  
  if ((transport->caps & caps) != caps)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Transport is not capable of channel-less peer-to-peer datagrams", 8);
    impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, value_local->s);
  }
  
  channel = impulse_channel_create(transport);
  if (impulse_channel_open(channel, value_local->s,
                              connect ? value_remote->s : NULL, 0) == -1)
  {
    impulse_unistring desc = impulse_unistring_create_utf("Cannot create peer channel", 8);
    impulse_error_raise(IMPULSE_ERROR_CRITICAL, desc, value_local->s);
  }
  
  impulse_channel_handle(channel, channel_event_handler, channel);
  impulse_event_register_fd(STDIN_FILENO, IMPULSE_EVENT_FD_READ, console_event_handler, NULL, channel);
  
  impulse_main();

  
  impulse_channel_close(channel);
  impulse_channel_destroy(channel);
  
  impulse_comms_cleanup();
  impulse_cleanup();
  
  return 0;
}
