/*                                                                  |- CFSFC -|
 * The contents of this file are subject to the Mozilla
 * Public License Version 1.0 (the "License"); you may
 * not use this file except in compliance with the
 * License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/ 
 *
 * Software distributed under the License is distributed
 * on an "AS IS" basis, WITHOUT WARRANTY OF
 * ANY KIND, either express or implied. See the
 * License for the specific language governing rights
 * and limitations under the License. 
 *
 * The Original Code is the ROWS widget library
 * for Impulse.
 *
 * The Initial Developer of the Original Code is
 * Andrew Hunter. Portions created by Andrew Hunter
 * are Copyright (C) 1999 Andrew Hunter.
 * All Rights Reserved
 *
 * Contributor(s):
 * No-one (yet :-)
 */

/*
 * main_xwin.c
 * -----------
 *
 * ROWS X-specific routines
 * Only include this file in places which are not architechture independant
 * Eventually the routines here will move to a seperate library.
 *
 * Routines here should only be accessed through the ROWS abstraction layer in
 * user programs.
 */

#include "impulse/rows/main_xwin.h"
#include "impulse/rows/input.h"
#include "impulse/rows/comms.h"

#include "impulse/comms/main.h"
#include "impulse/util/memory.h"
#include "impulse/util/main.h"
#include "impulse/util/event.h"

/*                            ======--------======                           */

ROWS_X_application appl = NULL;

static impulse_unistring opt_colourmap = NULL;
static impulse_unistring opt_display = NULL;
static impulse_unistring opt_sync = NULL;


/*                            ======--------======                           */

/* Error handling  */

#if USE_THREADS
static impulse_mutex_static error_mutex = IMPULSE_MUTEX_CREATE_STATIC;
#endif

static struct _err_info {
#if USE_THREADS
  impulse_thread owner;
#endif
  impulse_bool expecting_flag;
  impulse_bool error_flag;

  struct _err_info *next;
} *error_information;


impulse_bool ROWS_X_get_error_flag(void)
{
#if USE_THREADS
  impulse_thread our_thread;
  struct _err_info *info, *this_info;
  impulse_bool result;

  impulse_mutex_lock(&error_mutex);
  
  our_thread = impulse_thread_self();
  
  info = error_information;
  this_info = NULL;
  
  while (info)
  {
    if (impulse_thread_same(info->owner,our_thread))
      this_info = info;
    
    info = info->next;
  }

  if (this_info)
    result = this_info->error_flag;
  else
    result = 0;

  impulse_mutex_unlock(&error_mutex);

  return result;
#else
  if (!error_information)
    return 0;
  else
    return error_information->error_flag;
#endif
}


void ROWS_X_expecting_error(impulse_bool flag)
{
#if USE_THREADS
  impulse_thread our_thread;
  struct _err_info *info, *this_info;

  impulse_mutex_lock(&error_mutex);

  our_thread = impulse_thread_self();

  info = error_information;
  this_info = NULL;
  
  while (info)
  {
    if (impulse_thread_same(info->owner,our_thread))
      this_info = info;
    
    info = info->next;
  }
  
  if (!this_info)
  {
    this_info = impulse_alloc(sizeof(struct _err_info));
    
    this_info->owner          = our_thread;
    this_info->expecting_flag = 0;
    this_info->error_flag     = 0;
    this_info->next           = error_information;
    error_information         = this_info;
  }
  
  this_info->expecting_flag = flag;
  if (flag)
    this_info->error_flag = 0;
 
  impulse_mutex_unlock(&error_mutex);
#else
  if (!error_information)
  {
    error_information = impulse_alloc(sizeof(struct _err_info));

    error_information->expecting_flag = 0;
    error_information->error_flag     = 0;
    error_information->next           = NULL;
  }
  error_information->expecting_flag = flag;
  if (flag)
    error_information->error_flag = 0;
#endif
}


static impulse_int ROWS_X_error(Display     *display,
				XErrorEvent *error)
{
#if USE_THREADS
  impulse_thread    our_thread;
  struct _err_info *info, *this_info;
  impulse_int fall_through;

  impulse_mutex_lock(&error_mutex);

  our_thread = impulse_thread_self();

  info      = error_information;
  this_info = NULL;
  
  while (info)
  {
    if (impulse_thread_same(info->owner, our_thread))
      this_info = info;
    
    info = info->next;
  }

  fall_through = 0;

  if (this_info)
  {
    if (this_info->expecting_flag)
    {
      this_info->error_flag = error->error_code;
      fall_through = 1;
    }
  }

  if (!fall_through)
  {
    impulse_char erm[256];
    
    XGetErrorText(display, error->error_code, erm, 256);
    
    ROWS_error_raise(IMPULSE_ERROR_CRITICAL,
		     ROWS_ERROR_INTERNAL, 0, erm);
  }

  impulse_mutex_unlock(&error_mutex);

  return 0;
#else
  impulse_int fall_through = 0;
  
  if (error_information)
  {
    if (error_information->expecting_flag)
    {
      error_information->error_flag = error->error_code;
      fall_through = 1;
    }
  }
  
  if (!fall_through)
  {
    impulse_char erm[256];
    
    XGetErrorText(display, error->error_code, erm, 256);
    
    ROWS_error_raise(IMPULSE_ERROR_CRITICAL,
		     ROWS_ERROR_INTERNAL, 0, erm);
  }

  return 0;
#endif
}


/*                            ======--------======                           */

static impulse_bool fd_event_handler(impulse_event  *event,
				     void *data)
{
  /*
   * Most of the old problems with select, Xlib buffering and events
   * have gone away. The only problem is with multiple threads.
   * 
   * One thread waits on the X fd, and when it becomes ready it goes
   * off and calls this function. However in between finishing select
   * and calling this func, a second thread starts to select. Of course
   * this second thread calls the preselect hook which grabs all the
   * events the first thread was about to get hold of!
   * 
   * This can be solved by associating the preselect hook with the file
   * descriptor rather than globally. Hmmm, time to go work on impulse_event
   * me thinks.... DPB (5/12/1999)
   */
  
  /* 
   * Uncommenting this line, and the other in the preselect hook
   * illustrates what the above comment is talking about
   */
//#if 0
  if (!XPending(appl->display))
  {
//    printf("*** Arse - someone else stole my events %x\n", impulse_thread_self());
    return 1;
  }
//#endif

  while (XPending(appl->display))
  {
    XEvent         event;
    XEvent        *data;
    impulse_event  im_event;
    
    XNextEvent(appl->display, &event);      
    
    data  = impulse_alloc(sizeof(XEvent));
    *data = event;
    
    im_event.type = IMPULSE_EVENT_GUI;
    im_event.data.gui.data = data;
    im_event.data.gui.type = data->type;
    im_event.data.gui.object = data->xany.window;
    im_event.id = 0;
      
    impulse_event_trigger(&im_event);
  }
//  printf("Got some events %x\n", impulse_thread_self());
  return 1;
}


static impulse_bool event_preselect(void)
{
  if (!XPending(appl->display))
    return 0;
  
  do
  {
    XEvent         event;
    XEvent        *data;
    impulse_event  im_event;
    
    XNextEvent(appl->display, &event);
    
    data  = impulse_alloc(sizeof(XEvent));
    *data = event;
    
    im_event.type = IMPULSE_EVENT_GUI;
    im_event.data.gui.data = data;
    im_event.data.gui.type = data->type;
    im_event.data.gui.object = data->xany.window;
    im_event.id = 0;
    
    impulse_event_occured(&im_event);
  } while(XPending(appl->display));
  /* 
   * Uncomment this to see how the event stealing problem works
   */
//#if 0
//    printf("*** Drink - getting hold of some events %x\n", impulse_thread_self());
//#endif
  return 1;
}

/* Initialisation */

static void ROWS_after_init(void)
{
  impulse_command_value value;
  impulse_char *display = NULL;

  impulse_mutex_X_lock;

#if USE_THREADS && !GUI_XWIN_THREAD_PROTECT
  if (!XInitThreads())
    ROWS_error_raise(IMPULSE_ERROR_CRITICAL, ROWS_ERROR_INTERNAL,
		     0, "Xlib thread support failed to initialise");
#endif

  XSetErrorHandler(ROWS_X_error);

  if (impulse_command_set(opt_display, &value))
    display = impulse_unistring_to_utf(value[0].s, 8, 0);
  
  if (!(appl->display = XOpenDisplay(display)))
    ROWS_error_raise(IMPULSE_ERROR_CRITICAL, ROWS_ERROR_NO_GUI,
		     0, display);
  impulse_free(display);

  if (!(appl->visual = DefaultVisual(appl->display, DefaultScreen(appl->display))))
    ROWS_error_raise(IMPULSE_ERROR_WARNING, ROWS_ERROR_INTERNAL, 0,
		     "Can't get default visual");

  appl->screen_window = RootWindow(appl->display,
				   DefaultScreen(appl->display));

  if (impulse_command_set(opt_colourmap, NULL))
    appl->cmap = XCreateColormap(appl->display, appl->screen_window,
				 appl->visual, AllocNone);
  else
    appl->cmap = DefaultColormap(appl->display, DefaultScreen(appl->display));  
  appl->connection_fd = ConnectionNumber(appl->display);

  if (impulse_command_set(opt_sync, NULL))
    XSynchronize(appl->display, True);

  impulse_event_register_fd(appl->connection_fd,
                            IMPULSE_EVENT_FD_READ,
			    fd_event_handler,
			    event_preselect,
                            NULL);

  impulse_mutex_X_unlock;

  impulse_comms_init();
  ROWS_comms_init();
  
  impulse_comms_register(ROWS_comms_transport);
}


void ROWS_init(impulse_int argc,
	       impulse_char *argv[])
{
  impulse_unistring help;
  impulse_unistring arg;
  impulse_init(argc, argv);
  
  appl = impulse_alloc(sizeof(struct _ROWS_X_application));

#if USE_THREADS
# if GUI_XWIN_THREAD_PROTECT
  appl->xlib_lock = impulse_mutex_create();
# endif
  appl->fonts_lock = impulse_mutex_create();
  appl->colourmap_lock = impulse_mutex_create();
  appl->atoms_lock = impulse_mutex_create();
#endif

  appl->atoms = impulse_hash_create();
  appl->display = NULL;
  appl->visual = 0;
  appl->screen_window = 0;
  appl->fonts = impulse_hash_create();
  appl->cmap = 0;
  appl->connection_fd = -1;

  opt_colourmap = impulse_unistring_create_utf("colourmap", 8);
  help = impulse_unistring_create_utf("use a private colourmap", 8);
  impulse_command_option(opt_colourmap, 0, help);
  impulse_unistring_destroy(help);
  
  opt_display = impulse_unistring_create_utf("display", 8);
  arg = impulse_unistring_create_utf("name", 8);
  help = impulse_unistring_create_utf("connect to X server on display [name]", 8);
  impulse_command_option_arg(opt_display, 0, arg, IMPULSE_COMMAND_TYPE_STRING, help);
  impulse_unistring_destroy(arg);
  impulse_unistring_destroy(help);
  
  opt_sync = impulse_unistring_create_utf("sync", 8);
  help = impulse_unistring_create_utf("synchronise requests with X server", 8);
  impulse_command_option(opt_sync, 0, help);
  impulse_unistring_destroy(help);
  
  impulse_command_hook(ROWS_after_init);
}


void ROWS_cleanup(void)
{
  impulse_comms_deregister(ROWS_comms_transport);
  ROWS_comms_cleanup();
  impulse_comms_cleanup();
  
  impulse_hash_destroy(appl->atoms);
  impulse_hash_destroy(appl->fonts);
  
  XCloseDisplay(appl->display);
  
  impulse_free(appl);
  appl = NULL;  
}


Atom ROWS_X_atom_id(impulse_unistring name)
{
  Atom value = 0;
  
#if USE_THREADS
  impulse_mutex_lock(appl->atoms_lock);
#endif
  
  if (!(value = (Atom)impulse_hash_recover(appl->atoms,
				     name)))
  {
    impulse_char *cname = impulse_unistring_to_utf(name, 8, 0);
    impulse_mutex_X_lock;
    
    value = XInternAtom(appl->display, cname, False);
    
    impulse_mutex_X_unlock;
    impulse_hash_store(appl->atoms, name, (void *)value);
    impulse_free(cname);
  }
  
#if USE_THREADS
  impulse_mutex_unlock(appl->atoms_lock);
#endif
  
  return value;
}


impulse_int ROWS_X_get_button(impulse_int x_button)
{
  switch (x_button)
  {
   case Button1:
    return ROWS_MOUSE_BUTTON_LEFT;
    
   case Button2:
    return ROWS_MOUSE_BUTTON_MIDDLE;
    
   case Button3:
    return ROWS_MOUSE_BUTTON_RIGHT;
    
   case Button4:
    return ROWS_MOUSE_BUTTON_ROLLUP;
    
   case Button5:
    return ROWS_MOUSE_BUTTON_ROLLDOWN;
    
   default:
    return ROWS_MOUSE_BUTTON_LEFT;
  }
}


impulse_int ROWS_X_get_state(impulse_int x_state)
{
  impulse_int state = 0;
  
  if (x_state & Button1Mask)
    state |= ROWS_STATE_BUTTON_LEFT;
  if (x_state & Button2Mask)
    state |= ROWS_STATE_BUTTON_MIDDLE;
  if (x_state & Button3Mask)
    state |= ROWS_STATE_BUTTON_RIGHT;
  
  if (((x_state & ShiftMask) && !(x_state & LockMask)) ||
      (!(x_state & ShiftMask) && (x_state & LockMask)))
    state |= ROWS_STATE_KEY_SHIFT;
  if (x_state & Mod1Mask)
    state |= ROWS_STATE_KEY_ALT;
  if (x_state & ControlMask)
    state |= ROWS_STATE_KEY_CONTROL;

  return state;
}


/* End of file */
