#!/usr/bin/perl

#                                                                 |- Impulse -|
# The contents of this file are subject to the Mozilla
# Public License Version 1.0 (the "License"); you may
# not use this file except in compliance with the
# License. You may obtain a copy of the License at
# http://www.mozilla.org/MPL/ 
#
# Software distributed under the License is distributed
# on an "AS IS" basis, WITHOUT WARRANTY OF
# ANY KIND, either express or implied. See the
# License for the specific language governing rights
# and limitations under the License. 
#
# The Original Code is the ROWS widget library
# for Impulse.
#
# The Initial Developer of the Original Code is
# Andrew Hunter. Portions created by Andrew Hunter
# are Copyright (C) 1999 Andrew Hunter.
# All Rights Reserved
#
# Contributor(s):
# No-one (yet :-)
#

#
# Perl script for creating C files containing 8-bit to Unicode mapping
# descriptions.
#
# Syntax: mapping2c.pl <map file name> <c file name> <h file name> 
#           <mapping variable name> <mapping name>
#

#
# Uses the file format set down by the files supplied by Unicode, Inc. on the
# Unicode Standard 2.0 CDROM. That is:
#
# Lines beginning with a '#' are comments
# Empty lines (containing only whitespace) are ignored
# Other lines contain 3 fields, seperated by whitespace - 8-bit value &
#   Unicode value (in hexidecimal, prefixed by 0x)
#

use IO::Handle;
use File::Basename;
use POSIX;

$infile     = $ARGV[0];
$outcfile   = $ARGV[1];
$outhfile   = $ARGV[2];
$mapvarname = $ARGV[3];
$mapname    = $ARGV[4];

print STDERR "mapping2c.pl v0.1 written by Andrew Hunter\n";

# Read the data

open INFILE, "<$infile";

print STDERR "Reading mapping data from $infile.";
flush STDERR;

$chars = 0;

while (<INFILE>)
  {
    if (/^\s*$/)
      {
	# Whitespace. Continue
      }
    elsif (/^\s*\#(.*)$/)
      {
	# Comment. Continue
      }
    elsif (/^\s*0x([0-9A-Fa-f]+)\s+0x([0-9A-Fa-f]+)\s+(.*)$/)
      {
	# Character mapping
	$chars++;
	$eight = hex($1);
	$uni   = hex($2);

	if ($eight < 0   || $uni < 0 ||
	    $eight > 255 || $uni > 65535)
	  {
	    die "\nSyntax error at line $.\n";
	  }
	
	$eightuni[$eight] = $uni;
	$unieight{$uni}   = $eight;

	$occupied[floor($uni/256)] = 1;	   

	if (($chars%16) == 0)
	  {
	    print STDERR ".";
	    flush STDERR;
	  }
      }
    else
      {
	die "\nSyntax error at line $.\n";
      }
  }

print STDERR "\n";
print STDERR "Read $chars characters\n";

close INFILE;

# Output the header file

print STDERR "Generating header file...\n";

open HFILE, ">$outhfile";
print HFILE "/*\n * Header file automatically generated by mapping2c.pl\n * Mapping describes the $mapname mapping\n * This file is part of Impulse and comes under the Mozilla Public License\n * Data described in the variables defined here may be under differing\n * licensing conditions - see the original mapping file for full details\n *\n * mapping2c.pl was written by Andrew Hunter\n */\n\n";

($name, $path, $suffix) = fileparse($outhfile, [ ]);
$headerdef =  $name;
$headerdef =~ s/\./_/g;

print HFILE "#ifndef __modules_mapping_$headerdef\n#define __modules_mapping_$headerdef\n\n";

print HFILE "#include <impulse/util/unidata.h>\n\n";
print HFILE "#include <impulse/util/mapping.h>\n\n";
print HFILE "extern struct _impulse_mapping_to_8bit $mapvarname" . "_to_8bit;\n\n";
print HFILE "extern struct _impulse_mapping_from_8bit $mapvarname" . "_from_8bit;\n\n";

print HFILE "#endif\n";
close HFILE;

for ($char=0; $char<256; $char++)
  {
    if (!defined($eightuni[$char]))
      {
	# Control character/direct mapping
	$eightuni[$char] = $char;
	$unieight{$char} = $char;
	$occupied[0] = 1;
      }
  }

# Output the C file

print STDERR "Generating C file...\n";

open CFILE, ">$outcfile";
print CFILE "/*\n * C file automatically generated by mapping2c.pl\n * Mapping describes the $mapname mapping\n * The contents of this file may be based on data under differing license\n * agreements from the rest of libImpLib. See the original mapping file for \n * any relevant details\n *\n * mapping2c.pl was written by Andrew Hunter\n */\n\n";
print CFILE "#include <stdio.h>\n";
print CFILE "#include \"$outhfile\"\n\n";

# Unicode to 8-bit structures

print CFILE "static char unknown[] = \"unknown\";\n\n";
print CFILE "static struct _impulse_mapping_to_8bit_int U8_empty[256] =\n";
print CFILE "{\n";
for ($x=0; $x<256; $x++)
  {
    print CFILE "  {\n";
    print CFILE "    UNICODE_MAP_UNMAPPED,\n";
    print CFILE "    0,\n";
    print CFILE "    { 9, unknown }\n";
    
    if ($x<255)
      {
	print CFILE "  },\n";
      }
    else
      {
	print CFILE "  }\n";
      }
  }
print CFILE "};\n";

for ($section=0; $section<256; $section++)
  {
    if ($occupied[$section])
      {
	printf CFILE "static struct _impulse_mapping_to_8bit_int U8_%04x\[256] =\n", $section*256;
	print CFILE "{\n";
	$start = $section*256;
	$end   = $start+256;
	for ($char=$start; $char<$end; $char++)
	  {
	    print CFILE "  {\n";

	    if ($char < 32)
	      {
		# Control code
		$unieight{$char} = $char;
	      }

	    if (defined($unieight{$char}))
	      {
		print  CFILE "    UNICODE_MAP_SINGLE,\n";
		printf CFILE "    0x%02x,\n", $unieight{$char};
		print  CFILE "    { 0, NULL }\n";
	      }
	    else
	      {
		print CFILE "    UNICODE_MAP_UNMAPPED,\n";
		print CFILE "    0,\n";
		print CFILE "    { 9, unknown }\n";
	      }
	    
	    if ($char<($end-1))
	      {
		print CFILE "  },\n";
	      }
	    else
	      {
		print CFILE "  }\n";
	      }
	  }
	print CFILE "};\n";
      }
  }

# 8-bit to unicode data
print CFILE "\nstruct _impulse_mapping_from_8bit $mapvarname" . "_from_8bit = \n{\n";
print CFILE "  {\n";
for ($char=0; $char<256; $char++)
  {
    print CFILE "      {\n";

    print  CFILE "        UNICODE_MAP_SINGLE,\n";
    printf CFILE "        0x%04x,\n", $eightuni[$char];
    print  CFILE "        { 0, NULL }\n";

    if ($char < 255)
      {
	print CFILE "      },\n";
      }
    else
      {
	print CFILE "      }\n";
      }
  }
print CFILE "    }\n  };\n";

# Unicode to 8-bit data
print CFILE "\nstruct _impulse_mapping_to_8bit $mapvarname" . "_to_8bit = \n{\n";
print CFILE "    {\n      ";

for ($section=0;$section<256;$section++)
  {
    if ($occupied[$section] == 1)
      {
	printf CFILE "U8_%04x", $section*256;
      }
    else
      {
	printf CFILE "U8_empty";
      }

    if ($section<255)
      {
	print CFILE ", ";
	if ((($section+1)%8) == 0)
	  {
	    print CFILE "\n      ";
	  }
      }
    else
      {
	print CFILE "\n";
      }
  }

print CFILE "    }\n";

print CFILE "};\n";

close CFILE;

# Finish up

print STDERR "That seems to be it\n";
